open Compiler_meta;;
open Map_type;;
open Instructions;;
open Instructions.JVM;;
open Instructions.DEX;;
(*open Java_insn;;
open Dex_insn;;*)
open Handler;;
open Certificate;;

(* In the original dx tool, 7 corresponds to the SPECIAL_LABEL_COUNT (like *)
(* return block)                                                           *)
let getMinimumUnreservedLabel (maxLabel: int) = (maxLabel * 2) + 7;;

let getAvailableLabel (m: dBlockMap) (maxLabel: int) =
	let x = getMinimumUnreservedLabel (maxLabel) in
	let (maxKey, value) = Int32Map.max_binding (m) in
	let y = Int32.to_int maxKey in
	match x > y with
		true -> x
	| false -> y + 1
;;

let createInitBlock (insn: JVM_Instruction.t) (localN: int) (m: dBlockMap) 
: dBlockMap * bpMap * tsMap =
	let initialTS = Int32Map.add (0l) (localN) (Int32Map.empty) in
	match JVM_Instruction.args insn with
	(* to be precise there are instructions which are not possible to be the *)
	(* initial instruction. But we ignore it at this moment                  *)
		ArgsArgsType x ->
			let succOffset = (JVM_Instruction.label insn) + (JVM_Instruction.size insn) in
			(m,
				Int32Map.add (Int32.of_int succOffset) (0l) (Int32Map.add (0l) (0l) (Int32Map.empty)),
				Int32Map.add (Int32.of_int succOffset) (localN + 1) (initialTS))
	| GotoArgsType target_offset ->
			let succOffset = (JVM_Instruction.label insn) + target_offset in
			let startBlock = Block.add_succ (succOffset) (Block.update_pSucc (succOffset) (Block.empty)) in
			let targetBlock = Block.add_parent (0) (Block.empty) in
			let succOffset32 = Int32.of_int succOffset in
			(Int32Map.add (succOffset32) (targetBlock) (Int32Map.add (0l) (startBlock) (m)),
				Int32Map.add (succOffset32) (succOffset32) (Int32Map.add (0l) (0l) (Int32Map.empty)),
				Int32Map.add (succOffset32) (localN) (initialTS))
	| ClassArgsType x ->
			let succOffset = (JVM_Instruction.label insn) + (JVM_Instruction.size insn) in
			let startBlock = Block.add_succ (succOffset) (Block.update_pSucc (succOffset) (Block.empty)) in
			let targetBlock = Block.add_parent (0) (Block.empty) in
			let succOffset32 = Int32.of_int succOffset in
			(Int32Map.add (succOffset32) (targetBlock) (Int32Map.add (0l) (startBlock) (m)),
				Int32Map.add (succOffset32) (succOffset32) (Int32Map.add (0l) (0l) (Int32Map.empty)),
				Int32Map.add (succOffset32) (localN + 1) (initialTS))
	(* | FMI_insn x -> let succOffset = (x.label) + (x.size) in let          *)
	(* startBlock = {emptyBlock with succ=[succOffset]; psucc=(succOffset)}  *)
	(* in let targetBlock = {emptyBlock with parents=[0]} in (DBlockMap.add  *)
	(* (Int32.of_int succOffset) (targetBlock) (DBlockMap.add (0l)           *)
	(* (startBlock) (m)), BPMap.add (Int32.of_int succOffset) (Int32.of_int  *)
	(* succOffset) (BPMap.add (0l) (0l) (BPMap.empty)), TSMap.add            *)
	(* (Int32.of_int succOffset) (localN) (initialTS))                       *)
	| _ -> raise (Invalid_argument ("Bogus Initial Instruction " ^ JVM_Instruction.to_string insn));;

(* Still assume there is a return value *)
let rec createRetBlock (insns: JVM_Instruction.t list) (localN: int) (lvt: MethodPolicy.t) =
	match insns with
	(* [] -> raise (Invalid_argument "No Return Instruction") *)
		[] -> ( raise (Invalid_argument "No Return Instruction") )
	(* {emptyBlock with lastInsn=Some (create_ReturnDexInsn (-1) (1) (DEX_insn.Int))} *) (* This is only for matter of convenience to weaken the restriction that there must be a return instruction *)
	| h :: t -> match JVM_Instruction.args h with
				JVM.ReturnArgsType x ->
					let (dexRetType, flag) = match x with
							JVM.Void -> (DEX.Void, 0)
						| JVM.Int -> (DEX.Int, 1)
					in
					let typeInfo = JVM_Instruction.typeInfo h in
					let se = JVM_ins_typing.se typeInfo in
					let translated_rt = translate_st_rt (JVM_ins_typing.typing typeInfo) (localN) (lvt) (None) in
					let translated_rt = RT_op.update_flag (flag) (translated_rt) in
					  Block.add_insn (JVM_Instruction.label h, DEX.createReturnInsn (-1) (1) (dexRetType) (se) (translated_rt)) 
						(Block.update_typing (se) (translated_rt) Block.empty)
					(*{ emptyBlock with lastInsn = Some (create_ReturnDexInsn (-1) (1) (dexRetType)
									{ dex_se = x.jvm_typeInfo.jvm_se; rt = translated_rt })}*)
			| _ -> createRetBlock (t) (localN) (lvt);;

let visitOneStepInsn (insn:JVM_Instruction.t) (m: dBlockMap) (bp: bpMap) (ts: tsMap) (sb: sbMap)
		(isStartBlock: bool) (tsValue: int) =
	let il = JVM_Instruction.label insn in
	let is = JVM_Instruction.size insn in
	let isNextStart = (Int32Map.find (Int32.of_int (il + is)) sb) in
	let blockIndex = match isStartBlock with
			true -> Int32.of_int il
		| false -> Int32Map.find (Int32.of_int (il)) bp
	in
	let cb = Int32Map.find (blockIndex) m in
	match isNextStart with
		true ->
	(* Next instruction is start of another block *)
			(let succb = Int32Map.find (Int32.of_int (il + is)) m in
				(
					Int32Map.add (Int32.of_int (il + is))
					  (Block.add_parent (Int32.to_int blockIndex) succb)
						(*({ succb with parents = ((Int32.to_int blockIndex):: (succb.parents)) })*)
						(Int32Map.add (blockIndex)
						(Block.add_succ (il + is) (Block.update_pSucc (il+is) cb))
								(*{ cb with succ = ((il + is) :: cb.succ); psucc = (il + is) }*) m),
					Int32Map.add (Int32.of_int (il + is)) (Int32.of_int (il + is)) (bp),
					Int32Map.add (Int32.of_int (il + is)) (tsValue) (ts)
				))
	| false -> ((m, Int32Map.add (Int32.of_int (il + is)) (blockIndex) bp,
					Int32Map.add (Int32.of_int (il + is)) (tsValue) (ts)))
(* match isStartBlock with true -> (let cb = DBlockMap.find (Int32.of_int  *)
(* il) m in current instruction is start of the block match isNextStart    *)
(* with true -> Next instruction is start of another block (let succb =    *)
(* DBlockMap.find (Int32.of_int (il + is)) m in ( DBlockMap.add            *)
(* (Int32.of_int (il+is)) ({succb with parents=((il)::(succb.parents))})   *)
(* (DBlockMap.add (Int32.of_int il) ({cb with succ=((il+is) :: cb.succ);   *)
(* psucc=(il+is)}) m), bp, TSMap.add (Int32.of_int (il+is)) (tsValue) (ts) *)
(* )) | false -> ((m, BPMap.add (Int32.of_int (il + is)) (Int32.of_int     *)
(* (il)) bp, TSMap.add (Int32.of_int (il+is)) (tsValue) (ts))) Next        *)
(* instruction will be part of current block ) | false -> (let bi =        *)
(* BPMap.find (Int32.of_int (il)) (bp) in let cb = DBlockMap.find (bi) (m) *)
(* in match isNextStart with true -> Next instruction is start of another  *)
(* block (let succb = DBlockMap.find (Int32.of_int (il+is)) m in (         *)
(* DBlockMap.add (Int32.of_int (il+is)) ({succb with                       *)
(* parents=((Int32.to_int bi) :: succb.parents)}) (DBlockMap.add (bi) ({cb *)
(* with succ=((il+is) :: cb.succ); psucc=(il+is)}) m), bp, TSMap.add       *)
(* (Int32.of_int (il+is)) (tsValue) (ts) )) | false -> ((m, BPMap.add      *)
(* (Int32.of_int (il + is)) (bi) bp, TSMap.add (Int32.of_int (il+is))      *)
(* (tsValue) (ts))) Next instruction will be part of current block )       *)
;;

let visitGotoInsn (insn: JVM_Instruction.t) (m: dBlockMap) (bp: bpMap) (ts: tsMap) (sb: sbMap)
		(isStartBlock: bool) (tsValue: int) =
	let il = JVM_Instruction.label insn in
	let target_offset = (JVM.getGotoArgs insn) in
	let succ_idx32 = Int32.of_int (il + target_offset) in
	let blockIndex = match isStartBlock with
			true -> Int32.of_int il
		| false -> Int32Map.find (Int32.of_int (il)) bp
	in
	let cb = Int32Map.find (blockIndex) m in
	let succb = Int32Map.find (succ_idx32) m in
	(
		Int32Map.add (succ_idx32)
		  (Block.add_parent (Int32.to_int blockIndex) succb)
			(*{ succb with parents = ((Int32.to_int blockIndex):: (succb.parents)) }*) 
			(Int32Map.add (blockIndex)
			  (Block.add_succ (il + target_offset) (Block.update_pSucc (il+target_offset) cb))
					(*{ cb with succ = ((il + target_offset) :: cb.succ); psucc = (il + target_offset) }*) 
			m),
		Int32Map.add (succ_idx32) (succ_idx32) bp,
		Int32Map.add (succ_idx32) (tsValue) (ts)
	)
;;

let visitIfeqInsn (insn: JVM_Instruction.t) (m: dBlockMap) (bp: bpMap) (ts: tsMap) (sb: sbMap)
		(isStartBlock: bool) (tsValue: int) =
	let il = JVM_Instruction.label insn in
	let is = JVM_Instruction.size insn in
	let target_offset = JVM.IfArgs.offset (JVM.getIfArgs insn) in
	let succ1_idx32 = Int32.of_int (il + is) in
	let succ2_idx32 = Int32.of_int (il + target_offset) in
	let newBPMap = Int32Map.add (succ2_idx32) (succ2_idx32)
			(Int32Map.add (succ1_idx32) (succ1_idx32) bp) in
	let newTSMap = Int32Map.add (succ1_idx32) (tsValue) (Int32Map.add (succ2_idx32) (tsValue) (ts)) in
	let blockIndex = match isStartBlock with
			true -> Int32.of_int il
		| false -> Int32Map.find (Int32.of_int (il)) bp
	in
	let cb = Int32Map.find (blockIndex) m in
	let succt = Int32Map.find (succ2_idx32) m in
	let succf = Int32Map.find (succ1_idx32) m in
	(
		Int32Map.add (succ2_idx32)
		  (Block.add_parent (Int32.to_int blockIndex) (succt))
			(*{ succt with parents = ((Int32.to_int blockIndex):: (succt.parents)) }*) 
			(Int32Map.add (succ1_idx32)
			  (Block.add_parent (Int32.to_int blockIndex) (succf))
					(*{ succf with parents = ((Int32.to_int blockIndex):: (succf.parents)) }*) 
					(Int32Map.add (blockIndex)
					   (Block.add_succ (il+is) (Block.add_succ (il+target_offset) (Block.update_pSucc (il+is) cb)))
							(*{ cb with succ = ((il + is) :: ((il + target_offset) :: cb.succ)); psucc = (il + is) }*) 
		m)),
		newBPMap,
		newTSMap
	)
;;

(* Helper for addHandler to add all the handler block covering the current *)
(* pc as successor blocks                                                  *)
let rec addHandlers_rec (startBlock_pc: int32) (current: Block.t) (m: dBlockMap) (bp: bpMap)
		(ts: tsMap) (tsValue: int) (catches : Handler.t list) (maxLabel: int) =
	match catches with
		[] -> (m, bp, ts)
	| h :: t ->
			let handler_idx32 = Handler.handlerPC32 h in
			let intermediate_idx = (Handler.handlerPC h + maxLabel) in
			let intermediate_idx32 = Int32.of_int (intermediate_idx) in
			let intermediateBlock = Int32Map.find (intermediate_idx32) (m) in
			let handlerBlock = Int32Map.find (handler_idx32) (m) in
			let newHandler = Block.add_parent (intermediate_idx) (handlerBlock) in
			(*{ handlerBlock with parents = ((intermediate_idx):: handlerBlock.parents) } in*)
			let newCurrent = Block.add_succ (intermediate_idx) (current) in
				(*{ current with succ = ((intermediate_idx) :: current.succ) } in*)
			let newIntermediate = 
				Block.add_succ (Handler.handlerPC h) (Block.add_parent (Int32.to_int startBlock_pc)
				(Block.update_pSucc (Handler.handlerPC h) 
				(Block.add_insn (Handler.handlerPC h, DEX.createMoveExceptionInsn (-1) (1) (tsValue -1) (Level.bot_level) (RegistersType.empty)) 
				(intermediateBlock))))
				(*{ intermediateBlock with succ = (h.handlerPC :: intermediateBlock.succ);*)
				(*parents = ((Int32.to_int (startBlock_pc)) :: intermediateBlock.parents);*)
				(*psucc = (h.handlerPC); lastInsn = Some (create_ArgDexInsn (-1) (1) (MoveException)
							(tsValue - 1) ({ dex_se = Low; rt = Int32Map.empty })) }*) in
			(* TODO : type derivation for exceptions is definitely not correct *)
			let newM = Int32Map.add (handler_idx32) (newHandler)
					(Int32Map.add (startBlock_pc) (newCurrent)
							(Int32Map.add (intermediate_idx32) (newIntermediate) m))
			in
			let newBPMap = Int32Map.add (handler_idx32) (handler_idx32)
					(Int32Map.add (intermediate_idx32) (intermediate_idx32) (bp))
			in
			let newTSMap = Int32Map.add (handler_idx32) (tsValue)
					(Int32Map.add (intermediate_idx32) (tsValue + 1) (ts))
			in
			addHandlers_rec (startBlock_pc) (newCurrent) (newM) (newBPMap) (newTSMap) (tsValue) (t) (maxLabel)

let addHandlers (startBlock_pc: int32) (pc: int) (m: dBlockMap) (bp: bpMap) (ts: tsMap)
		(tsValue: int) (catches : Handler.t list) (maxLabel: int) =
	let filteredHandlers = Handler.handlers_for (catches) (pc) in
	let current = Int32Map.find (startBlock_pc) m in
	let (m, bp, ts) = addHandlers_rec (startBlock_pc) (current) (m) (bp)(ts) (tsValue) (filteredHandlers) (maxLabel) in
	(m, bp, ts, filteredHandlers)

let withIntermediateBlock (m: dBlockMap) (bp: bpMap) (ts: tsMap) (parent_lbl: int) (succ_lbl: int) (tsValue: int) (maxLabel: int) 
(ins: (int*DEX_Instruction.t) option) (applicableHandlers : Handler.t list) =
	let newLabel = getAvailableLabel (m) (maxLabel) in
	let blockWithIns = match ins with 
	| Some (source, insn) -> Block.add_insn (source, insn) (Block.empty)
	| None -> Block.empty in
	let newBlock = Block.add_parent (parent_lbl) (Block.add_succ (succ_lbl) 
	  (Block.update_pSucc (succ_lbl) (Block.update_handlers (applicableHandlers) (blockWithIns)))) in
		(*{ emptyBlock with parents =[parent_lbl]; succ =[succ_lbl]; psucc = succ_lbl; lastInsn = ins; catches = applicableHandlers } in*)
	let currBlock = Int32Map.find (Int32.of_int parent_lbl) (m) in
	let succBlock = Int32Map.find (Int32.of_int succ_lbl) (m) in
	let currBlock = Block.add_succ (newLabel) (Block.update_pSucc newLabel 
	  (Block.update_handlers applicableHandlers currBlock)) in
	(*{ currBlock with succ = (newLabel :: currBlock.succ); psucc = newLabel; catches = applicableHandlers } in*)
	let succBlock = Block.add_parent (newLabel) (succBlock) in
	(*{ succBlock with parents = (newLabel :: succBlock.parents) } in*)
	(
		Int32Map.add (Int32.of_int (newLabel)) (newBlock)
			(Int32Map.add (Int32.of_int (parent_lbl)) (currBlock)
					(Int32Map.add (Int32.of_int (succ_lbl)) (succBlock) (m))),
		Int32Map.add (Int32.of_int (newLabel)) (Int32.of_int (newLabel))
			(Int32Map.add (Int32.of_int (succ_lbl)) (Int32.of_int (succ_lbl)) (bp)),
		(* SBMap.add (Int32.of_int (newLabel)) (true) (SBMap.add (Int32.of_int *)
		(* (succ_lbl)) (true) (sb)),                                           *)
		Int32Map.add (Int32.of_int (newLabel)) (tsValue - 1)
			(Int32Map.add (Int32.of_int (succ_lbl)) (tsValue) (ts))
	)
;;

let visitThrowingInsn (insn:JVM_Instruction.t) (m: dBlockMap) (bp: bpMap) (ts: tsMap)
		(isStartBlock: bool) (tsValue: int) (catches: Handler.t list) (localN: int) (maxLabel: int)
		(ins: (int * DEX_Instruction.t) option) =
	let il = JVM_Instruction.label insn in
	let is = JVM_Instruction.size insn in
	(* let succ_idx32 = Int32.of_int (il + is) in let newBPMap = BPMap.add   *)
	(* (succ_idx32) (succ_idx32) (bp) in let newTSMap = TSMap.add            *)
	(* (succ_idx32) (tsValue) (ts) in                                        *)
	let blockIndex = match isStartBlock with
			true -> Int32.of_int il
		| false -> Int32Map.find (Int32.of_int (il)) bp
	in
	(* let cb = DBlockMap.find (blockIndex) m in let succ = DBlockMap.find   *)
	(* (succ_idx32) m in let newM = DBlockMap.add (succ_idx32) ({succ with   *)
	(* parents=((Int32.to_int blockIndex)::(succ.parents))}) (DBlockMap.add  *)
	(* (blockIndex) ({cb with succ=((il+is) :: cb.succ); psucc=(il+is)}) m)  *)
	(* in                                                                    *)
	let (m, bp, ts, applicableHandlers) = addHandlers (blockIndex) (il) (m) (bp) (ts) (localN +1) 
	  (catches) (maxLabel) in
	withIntermediateBlock (m) (bp) (ts) (Int32.to_int blockIndex) (il + is) (tsValue) (maxLabel) 
	  (ins) (applicableHandlers)
(* newM, newBPMap, newTSMap *)
;;

let visitReturnInsn (insn: JVM_Instruction.t) (maxLabel: int) (m: dBlockMap) (bp: bpMap) (ts: tsMap) (sb: sbMap) 
(isStartBlock: bool) (retBlock: Block.t) =
	let il = JVM_Instruction.label insn in
	let is = JVM_Instruction.size insn in
	let blockIndex = match isStartBlock with
			true -> Int32.of_int il
		| false -> Int32Map.find (Int32.of_int (il)) bp
	in
	let cb = Int32Map.find (blockIndex) m in
	let newBPMap =
		if ((il + is) < maxLabel) then
			(let next_idx32 = Int32.of_int (il + is) in
				Int32Map.add (next_idx32) (next_idx32) bp)
		else
			(bp)
	in
	((
			Int32Map.add (blockIndex) (Block.add_succ (Block.retLabel) (Block.update_pSucc (Block.retLabel) (cb)))
			(*{ cb with succ = ((retLabel) :: cb.succ); psucc = (retLabel) }*) m,
			newBPMap,
			ts
		), 
		Block.add_parent (Int32.to_int blockIndex) (retBlock))
		(*{ retBlock with parents = ((Int32.to_int blockIndex) :: retBlock.parents) })*)
;;

let rec countArgs (str: string) (len: int) (count: int) =
	match str.[0] with
		'B' | 'C' | 'D' | 'F' | 'I' | 'J' | 'S' | 'Z' ->
			(match (len - 1) <= 0 with
					true -> (count + 1)
				| false -> countArgs (String.sub (str) (1) (len - 1)) (len - 1) (count + 1)
			)
	| '[' -> countArgs (String.sub (str) (1) (len - 1)) (len - 1) (count)
	| 'L' -> (let nextIndex = (String.index (str) (';')) + 1 in
				match nextIndex >= len with
					true -> (count + 1)
				| false -> countArgs (String.sub (str) (nextIndex) (len - nextIndex)) (len - nextIndex) (count + 1)
			)
	| _ as chr -> raise (Invalid_argument ("Invalid character for a method descriptor : " ^ Char.escaped (chr)))
;;

let parseInsn (insn: JVM_Instruction.t) (maxLabel: int) (localN: int) (m: dBlockMap) (bp: bpMap)
		(ts: tsMap) (sb: sbMap) (ret: Block.t) (catches: Handler.t list) 
		(oNextIns: JVM_Instruction.t option) (lvt: Certificate.MethodPolicy.t) =
	let currentTS = Int32Map.find (JVM_Instruction.label32 insn) (ts) in
	let isStartBlock = Int32Map.find (JVM_Instruction.label32 insn) (sb) in
	match JVM_Instruction.code insn with
	| JVM.Nop -> (visitOneStepInsn (insn) (m) (bp) (ts) (sb) (isStartBlock) (currentTS), ret)
	| Dup | Swap | Pop -> 
		          ( let newTS = match JVM_Instruction.code insn with
									Pop -> currentTS - 1
								| Swap -> currentTS
								| Dup -> currentTS + 1
								| _ as c -> raise (Invalid_argument ("Bogus opcode for Plain instruction" ^ JVM.string_of_insnType (c)))
							in
							(visitOneStepInsn (insn) (m) (bp) (ts) (sb) (isStartBlock) (newTS), ret))
	| JVM.Return ->
			(	let isStartBlock = (Int32Map.find (JVM_Instruction.label32 insn) sb) in
				visitReturnInsn (insn) (maxLabel) (m) (bp) (ts) (sb) (isStartBlock) (ret))
	| JVM.Throw -> (let blockIndex = (match isStartBlock with
										true -> JVM_Instruction.label32 insn
									| false -> Int32Map.find (Int32.of_int (JVM_Instruction.label insn)) bp)
							in
							let (m, bp, ts, applicableHandlers) = (addHandlers (blockIndex) (JVM_Instruction.label insn) 
							  (m) (bp) (ts) (localN + 1) (catches) (maxLabel)) in
							let currentBlock = Int32Map.find (blockIndex) (m) in
							let m = Int32Map.add (blockIndex) (Block.update_handlers (applicableHandlers) (currentBlock)) (m) in
							((m, bp, ts), ret) )
	| Store | Load | Push | IInc -> 
				(let newTS = match JVM_Instruction.code insn with
						Push
					| Load -> currentTS + 1
					| Store -> currentTS - 1
					| IInc -> currentTS
					| _ as c -> raise (Invalid_argument ("Bogus opcode for Args instruction " ^ JVM.string_of_insnType (c)))
				in
				(visitOneStepInsn (insn) (m) (bp) (ts) (sb) (isStartBlock) (newTS), ret)
			)
	| JVM.Binop -> (visitOneStepInsn (insn) (m) (bp) (ts) (sb) (isStartBlock) (currentTS - 1), ret) 
	| JVM.Goto -> (visitGotoInsn (insn) (m) (bp) (ts) (sb) (isStartBlock) (currentTS), ret) 
	| JVM.If -> (let ifArg = JVM.getIfArgs insn in
				let tsValue = match JVM.IfArgs.kind ifArg with
						JVM.Eq | JVM.Ne
					| JVM.Gt | JVM.Ge
					| JVM.Lt | JVM.Le -> (currentTS - 2)
					| JVM.Eqz | JVM.Nez
					| JVM.Gtz | JVM.Gez
					| JVM.Ltz | JVM.Lez -> (currentTS - 1)
				in
				(visitIfeqInsn (insn) (m) (bp) (ts) (sb) (isStartBlock) (tsValue), ret) )
	| JVM.New -> (visitThrowingInsn (insn) (m) (bp) (ts) (isStartBlock) (currentTS + 1) (catches) 
	    (localN) (maxLabel) (None), ret)
	| JVM.Getfield -> (visitThrowingInsn (insn) (m) (bp) (ts) (isStartBlock) (currentTS) (catches) 
	    (localN) (maxLabel) (None), ret)
	| JVM.Putfield -> (visitThrowingInsn (insn) (m) (bp) (ts) (isStartBlock) (currentTS - 2) (catches) 
	    (localN) (maxLabel) (None), ret)
	| JVM.Invoke t -> let arg = JVM.getFMIArgs insn in
		        let (tsValue) = (let closeIndex = String.index (FMIArgs.fmi_type arg) (')') in
								let str = String.sub (FMIArgs.fmi_type arg) (1) (closeIndex - 1) in
								let isVoid = (match t with JVM.Void -> 1 | JVM.Int -> 0) in
								match (String.length (str)) = 0 with
									false -> (currentTS - (countArgs (str) (String.length (str)) (0)) - isVoid)
								| true -> (currentTS - isVoid)) in
						let intermediateIns = 
							( let retSig = (FMIArgs.fmi_type arg).[(String.index (FMIArgs.fmi_type arg) (')')) + 1] in
								let noReturn = (retSig = 'V') in
								match noReturn with
									true -> None
								| false -> let Some nextInsn = oNextIns in 
									  let typeInfo = JVM_Instruction.typeInfo nextInsn in
	                  let current_se = JVM_ins_typing.se typeInfo in
										let st = JVM_ins_typing.typing typeInfo in
										let ([res], st') = ST_op.pop_n st 1 in
	                  let translated_rt = translate_st_rt (st') (localN) (lvt) (Some res) in
	                Some (JVM_Instruction.label insn, DEX.createMoveResultInsn (-1) (1) 
								        (tsValue - 1) (current_se) (translated_rt))
								(* TODO : need to fix this one for method instruction *)
							)
						in
				(visitThrowingInsn (insn) (m) (bp) (ts) (isStartBlock) (tsValue) (catches) (localN) (maxLabel)
						(intermediateIns), ret)
(*							
		Plain_insn ins ->
			(
				match ins.code with
					Java_insn.Throw ->
						(let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
							let blockIndex = (match isStartBlock with
										true -> Int32.of_int ins.label
									| false -> Int32Map.find (Int32.of_int (ins.label)) bp)
							in
							let (m, bp, ts, applicableHandlers) = (addHandlers (blockIndex) (ins.label) (m) (bp) (ts) (localN + 1) (catches) (maxLabel)) in
							let currentBlock = Int32Map.find (blockIndex) (m) in
							let m = Int32Map.add (blockIndex) ({ currentBlock with catches = applicableHandlers }) (m) in
							((m, bp, ts), ret) )
				| _ ->
						( let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
							let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
							let newTS = match ins.code with
									Pop -> currentTS - 1
								| Swap -> currentTS
								| Dup -> currentTS + 1
								| _ as c -> raise (Invalid_argument ("Bogus opcode for Plain instruction" ^ string_of_insnType (c)))
							in
							(visitOneStepInsn (ins) (m) (bp) (ts) (sb) (isStartBlock) (newTS), ret)
						)
			)
	| Args_insn ins ->
			( let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
				let newTS = match ins.code with
						Push
					| Load -> currentTS + 1
					| Store -> currentTS - 1
					| IInc -> currentTS
					| _ as c -> raise (Invalid_argument ("Bogus opcode for Args instruction " ^ string_of_insnType (c)))
				in
				(visitOneStepInsn (ins) (m) (bp) (ts) (sb) (isStartBlock) (newTS), ret)
			)
	| Binop_insn ins ->
			(
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
				(visitOneStepInsn (ins) (m) (bp) (ts) (sb) (isStartBlock) (currentTS - 1), ret)
			)
	| Goto_insn ins ->
			(
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
				(visitGotoInsn (ins) (m) (bp) (ts) (sb) (isStartBlock) (currentTS), ret)
			)
	| If_insn ins ->
			(
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
				let tsValue =
					match ins.args.kind with
						Java_insn.Eq | Java_insn.Ne
					| Java_insn.Gt | Java_insn.Ge
					| Java_insn.Lt | Java_insn.Le -> (currentTS - 2)
					| Java_insn.Eqz | Java_insn.Nez
					| Java_insn.Gtz | Java_insn.Gez
					| Java_insn.Ltz | Java_insn.Lez -> (currentTS - 1)
				in
				(visitIfeqInsn (ins) (m) (bp) (ts) (sb) (isStartBlock) (tsValue), ret)
			)
	| Return_insn ins ->
			(
				let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
				visitReturnInsn (ins) (maxLabel) (m) (bp) (ts) (sb) (isStartBlock) (ret)
			)
	| New_insn ins ->
			(
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
				let tsValue = currentTS + 1 in
				((visitThrowingInsn (ins) (m) (bp) (ts) (isStartBlock) (tsValue) (catches) (localN) (maxLabel) (None)), ret)
			)
	| FMI_insn ins ->
			(
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let isStartBlock = (Int32Map.mem (Int32.of_int ins.label) sb) in
				let (tsValue) =
					match ins.code with
						Java_insn.Getfield -> (currentTS)
					| Java_insn.Putfield -> (currentTS - 2)
					| Java_insn.Invoke t -> (let closeIndex = String.index (ins.args.fmi_type) (')') in
								let str = String.sub (ins.args.fmi_type) (1) (closeIndex - 1) in
								let isVoid = (match t with Java_insn.Void -> 1 | Java_insn.Int -> 0) in
								match (String.length (str)) = 0 with
									false -> (currentTS - (countArgs (str) (String.length (str)) (0)) - isVoid)
								| true -> (currentTS - isVoid))
					| _ -> raise (Invalid_argument "Invalid code for FMI instruction")
				in
				let (intermediateIns) = match ins.code with
						Java_insn.Invoke t ->
							( let retSig = ins.args.fmi_type.[(String.index (ins.args.fmi_type) (')')) + 1] in
								let noReturn = (retSig = 'V') in
								match noReturn with
									true -> None
								| false -> Some (create_ArgDexInsn (-1) (1) (MoveResult) (tsValue - 1)
												({ dex_se = Low; rt = Int32Map.empty }))
								(* TODO : need to fix this one for method instruction *)
							)
					| _ -> None
				in
				(visitThrowingInsn (ins) (m) (bp) (ts) (isStartBlock) (tsValue) (catches) (localN) (maxLabel)
						(intermediateIns), ret)
			)
*)
;; (*| _ -> raise (Invalid_argument "Under development");;*)

let rec parseInsns (jProg: JVM_Instruction.t list) (maxLabel: int) (localN: int) (m: dBlockMap)
		(bp: bpMap) (ts: tsMap) (sb: sbMap) (ret: Block.t) (catches: Handler.t list) 
		(lvt: Certificate.MethodPolicy.t) =
	match jProg with
		[] -> (m, bp, ts, ret)
	| insn :: t -> 
		  let oNextInsn = match t with
			  | [] -> None
			  | nextIns :: t' -> Some nextIns in
			let ((newM, newBP, newTS), ret) = 
				parseInsn (insn) (maxLabel) (localN) (m) (bp) (ts) (sb) (ret) (catches) (oNextInsn) (lvt)
			in 
			parseInsns (t) (maxLabel) (localN) (newM) (newBP) (newTS) (sb) (ret) (catches) (lvt);;

let parse (jProg: JVM_Instruction.t list) (maxLabel: int) (localN: int) (lvt: Certificate.MethodPolicy.t) 
    (m: dBlockMap) (sb: sbMap) (catches: Handler.t list) =
	let (m, bp, ts) = createInitBlock (List.hd jProg) (localN) (m) in
	let ret = createRetBlock (jProg) (localN) (lvt) in
	parseInsns (List.tl jProg) (maxLabel: int) (localN) (m) (bp) (ts) (sb) (ret) (catches) (lvt);;

